/* -*- coding: cp850-dos -*-  */
/* This file provides routines to convert a string from one of cp437, cp850
 * or ascii to iso-latin-1 and vice-versa.  This does unfortunately not mean
 * that it is possible to convert between the two DOS codepages with iso-latin
 * as an intermediate format, as most DOS graphics characters are absent from
 * iso-latin.  However, as far as pure text is concerned, the conversion is
 * possible.
 *
 * Btw, it's called cconvert as it converts between different character-sets.
 */

#include <stdio.h>
#include "cconvert.h"
#include <stdlib.h>
#include <string.h>

char cp_name[NUM_CP][10]={{"cp437"},{"latin1"},{"cp850"},{"us-ascii"}};

/* to can be NULL, as realloc will be used to alloc memory as needed.
 * The function returns a pointer to the (possibly new) location of
 * to, allocated with realloc() --- so *to should be either NULL or
 * point to a location previously allocated with malloc() or one of
 * its relatives... */

static unsigned char to_latin1_table[4][256] = {{0}};
static char from_latin1_table[4][256][3] = {{{0}}};
void init_cconvert(void);


char *cconvert(char*  from,
	       int    from_coding_system,
	       char** to,
	       int    to_coding_system)
{
  char *s = NULL;
  char *t = NULL;
  int n,i;

  /* first time called? */
  if (to_latin1_table[0][1] == 0) {
    init_cconvert();
  }

  /* conversion from cp437 or cp850 to latin1: */
  /* ascii is identical to latin1 for all characters used, so no conversion 
     is needed */
  if ((from_coding_system != latin1) && (from_coding_system != ascii)) {
    n = strlen(from);
    s = malloc(n+1);
    for (i=0; i<n; i++) {
      s[i] = to_latin1_table[from_coding_system][(unsigned char)from[i]];
    }
    s[n] = 0;
  } else {
    s = from;
  }
  
  /* conversion from latin1 to cp437, cp850, or ascii: */
  if (to_coding_system != latin1) {
    n = strlen(s);
    t = malloc(2*n+1);                 /* worst case */
    t[0] = 0;
    for (i=0; i<n; i++) {
      sprintf(&(t[strlen(t)]), "%s", from_latin1_table[to_coding_system][(unsigned char)s[i]]);
    }
    n = strlen(t);
    *to = realloc(*to, n+1);
    strncpy(*to, t, n);
    (*to)[n] = 0;
    if (s!=from) {free(s);}
    free(t);
  } else {                             /* target was latin1 */
    if (s != from) {
      free(*to);
      *to = s;
    } else {
      n = strlen(s);
      *to = realloc(*to, n+1);
      strncpy(*to, s, n);
      (*to)[n] = 0;
    }
  }
  recode(*to, strlen(*to)+1, to_coding_system);
  return *to;
}

void init_cconvert()
{
  int i,j;

  for (j=0; j<4; j++) {
    /* No codepage should use the first 32 characters: */
    for (i=0; i<32; i++) {
      to_latin1_table[j][i] = '?';
    }
    /* The characters 32--126 should be the same everywhere: */
    for (i=32; i<127; i++) {
      to_latin1_table[j][i] = i;
    }
    /* character 127 should be unused: */
    to_latin1_table[j][127] = '?';
  }
  /* the rest needs to be set individually */
  /* latin1 is of course the identity mapping */
  for (i=0; i<256; i++) {
    to_latin1_table[latin1][i] = i;
  }
  /* ascii knows no characters above 127 */
  for (i=128; i<256; i++) {
    to_latin1_table[ascii][i] = '?';
  }
  /* cp437 only knows a handfull of characters: */
  for (i=128; i<256; i++) {
    to_latin1_table[cp437][i] = '?';
  }
  to_latin1_table[cp437][0x80] = 0xC7;
  to_latin1_table[cp437][0x81] = 0xFC;
  to_latin1_table[cp437][0x82] = 0xE9;
  to_latin1_table[cp437][0x83] = 0xE2;
  to_latin1_table[cp437][0x84] = 0xE4;
  to_latin1_table[cp437][0x85] = 0xE0;
  to_latin1_table[cp437][0x86] = 0xE5;
  to_latin1_table[cp437][0x87] = 0xE7;
  to_latin1_table[cp437][0x88] = 0xEA;
  to_latin1_table[cp437][0x89] = 0xEB;
  to_latin1_table[cp437][0x8A] = 0xE8;
  to_latin1_table[cp437][0x8B] = 0xEF;
  to_latin1_table[cp437][0x8C] = 0xEE;
  to_latin1_table[cp437][0x8D] = 0xEC;
  to_latin1_table[cp437][0x8E] = 0xC4;
  to_latin1_table[cp437][0x8F] = 0xC5;
  to_latin1_table[cp437][0x90] = 0xC9;
  to_latin1_table[cp437][0x91] = 0xE6;
  to_latin1_table[cp437][0x92] = 0xC6;
  to_latin1_table[cp437][0x93] = 0xF4;
  to_latin1_table[cp437][0x94] = 0xF6;
  to_latin1_table[cp437][0x95] = 0xF2;
  to_latin1_table[cp437][0x96] = 0xFB;
  to_latin1_table[cp437][0x97] = 0xF9;
  to_latin1_table[cp437][0x98] = 0xFF;
  to_latin1_table[cp437][0x99] = 0xD6;
  to_latin1_table[cp437][0x9A] = 0xDC;
  to_latin1_table[cp437][0x9B] = 0xA2;
  to_latin1_table[cp437][0x9C] = 0xA3;
  to_latin1_table[cp437][0x9D] = 0xA5;
  to_latin1_table[cp437][0x9E] = '?';  /* Pts */
  to_latin1_table[cp437][0x9F] = '?';  /* long-s */
  to_latin1_table[cp437][0xA0] = 0xE1;
  to_latin1_table[cp437][0xA1] = 0xED;
  to_latin1_table[cp437][0xA2] = 0xF3;
  to_latin1_table[cp437][0xA3] = 0xFA;
  to_latin1_table[cp437][0xA4] = 0xF1;
  to_latin1_table[cp437][0xA5] = 0xD1;
  to_latin1_table[cp437][0xA6] = 0xAA;
  to_latin1_table[cp437][0xA7] = 0xBA;
  to_latin1_table[cp437][0xA8] = 0xBF;
  to_latin1_table[cp437][0xA9] = '?';  /* inverse not */
  to_latin1_table[cp437][0xAA] = 0xAC;
  to_latin1_table[cp437][0xAB] = 0xBD;
  to_latin1_table[cp437][0xAC] = 0xBC;
  to_latin1_table[cp437][0xAD] = 0xA1;
  to_latin1_table[cp437][0xAE] = 0xAB;
  to_latin1_table[cp437][0xAF] = 0xBB;
  to_latin1_table[cp437][0xE0] = '?';  /* alpha */
  to_latin1_table[cp437][0xE1] = 0xDF;
  to_latin1_table[cp437][0xE2] = '?';   /* right angle */
  to_latin1_table[cp437][0xE3] = '?';   /* pi */
  to_latin1_table[cp437][0xE4] = '?';   /* sum */
  to_latin1_table[cp437][0xE5] = '?';   /* sigma */
  to_latin1_table[cp437][0xE6] = 0xB5;
  to_latin1_table[cp437][0xE7] = '?';   /* tau */
  to_latin1_table[cp437][0xE8] = '?';   /* ??? */
  to_latin1_table[cp437][0xE9] = '?';   /* Sigma */
  to_latin1_table[cp437][0xEA] = '?';   /* Omega */
  to_latin1_table[cp437][0xEB] = '?';   /* delta */
  to_latin1_table[cp437][0xEC] = '?';   /* infinity */
  to_latin1_table[cp437][0xED] = 0xF8;
  to_latin1_table[cp437][0xEE] = '?';   /* epsilon */
  to_latin1_table[cp437][0xEF] = '?';   /* nu */
  to_latin1_table[cp437][0xF0] = '?';   /* identity */
  to_latin1_table[cp437][0xF1] = 0xB1;
  to_latin1_table[cp437][0xF2] = '?';   /* geq */
  to_latin1_table[cp437][0xF3] = '?';   /* leq */
  to_latin1_table[cp437][0xF4] = '?';   /* integral top */
  to_latin1_table[cp437][0xF5] = '?';   /* integral bottom */
  to_latin1_table[cp437][0xF6] = 0xF7;
  to_latin1_table[cp437][0xF7] = '?';   /* approx */
  to_latin1_table[cp437][0xF8] = 0xB0;
  to_latin1_table[cp437][0xF9] = 0xB7;
  to_latin1_table[cp437][0xFA] = 0xB7;
  to_latin1_table[cp437][0xFB] = '?';   /* tick */
  to_latin1_table[cp437][0xFC] = '?';   /* to n */
  to_latin1_table[cp437][0xFD] = 0xB2;
  to_latin1_table[cp437][0xFE] = '?';   /* square */
  /* cp850 knows many more, but uses the same positions for most: */
  for (i=128; i<256; i++) {
    to_latin1_table[cp850][i] = to_latin1_table[cp437][i];
  }
  to_latin1_table[cp850][0x9B] = 0xF8;
  to_latin1_table[cp850][0x9C] = 0xA3;
  to_latin1_table[cp850][0x9D] = 0xD8;
  to_latin1_table[cp850][0x9E] = 0xD7;
  to_latin1_table[cp850][0x9F] = '?';  /* long-s */
  to_latin1_table[cp850][0xA9] = 0xAE;
  to_latin1_table[cp850][0xB5] = 0xC1;
  to_latin1_table[cp850][0xB6] = 0xC2;
  to_latin1_table[cp850][0xB7] = 0xC0;
  to_latin1_table[cp850][0xB8] = 0xA9;
  to_latin1_table[cp850][0xBD] = 0xA2;
  to_latin1_table[cp850][0xBE] = 0xA5;
  to_latin1_table[cp850][0xC6] = 0xE3;
  to_latin1_table[cp850][0xC7] = 0xC3;
  to_latin1_table[cp850][0xCF] = 0xA4;
  to_latin1_table[cp850][0xD0] = 0xF0;
  to_latin1_table[cp850][0xD1] = 0xD0;
  to_latin1_table[cp850][0xD2] = 0xCA;
  to_latin1_table[cp850][0xD3] = 0xCB;
  to_latin1_table[cp850][0xD4] = 0xC8;
  to_latin1_table[cp850][0xD5] = '?';  /* dotless-i */
  to_latin1_table[cp850][0xD6] = 0xCD;
  to_latin1_table[cp850][0xD7] = 0xCE;
  to_latin1_table[cp850][0xD8] = 0xCF;
  to_latin1_table[cp850][0xDD] = 0xA6;
  to_latin1_table[cp850][0xDE] = 0xCC;
  to_latin1_table[cp850][0xE0] = 0xD3;
  to_latin1_table[cp850][0xE1] = 0xDF;
  to_latin1_table[cp850][0xE2] = 0xD4;
  to_latin1_table[cp850][0xE3] = 0xD2;
  to_latin1_table[cp850][0xE4] = 0xF5;
  to_latin1_table[cp850][0xE5] = 0xD5;
  to_latin1_table[cp850][0xE7] = 0xFE;
  to_latin1_table[cp850][0xE8] = 0xDE;
  to_latin1_table[cp850][0xE9] = 0xDA;
  to_latin1_table[cp850][0xEA] = 0xDB;
  to_latin1_table[cp850][0xEB] = 0xD9;
  to_latin1_table[cp850][0xEC] = 0xFD;
  to_latin1_table[cp850][0xED] = 0xDD;
  to_latin1_table[cp850][0xEE] = '?';  /* bar? */
  to_latin1_table[cp850][0xEF] = 0xB4;
  to_latin1_table[cp850][0xF0] = 0xAD;
  to_latin1_table[cp850][0xF2] = '?';  /* low equal */
  to_latin1_table[cp850][0xF3] = 0xBE;
  to_latin1_table[cp850][0xF4] = 0xB6;
  to_latin1_table[cp850][0xF5] = 0xA7;
  to_latin1_table[cp850][0xF7] = 0xB8;
  to_latin1_table[cp850][0xF9] = 0xA8;
  to_latin1_table[cp850][0xFB] = 0xB9;
  to_latin1_table[cp850][0xFC] = 0xB3;

  /* latin1 is the identity mapping */
  for (i=0; i<256; i++) {
    sprintf(from_latin1_table[latin1][i], "%c", i);
  }
  /* cp850 at least maps character for character */
  for (i=0; i<256; i++) {
    sprintf(from_latin1_table[cp850][i], "?");
  }
  for (i=0; i<256; i++) {
    if (to_latin1_table[cp850][i] != '?') {
      sprintf(from_latin1_table[cp850][to_latin1_table[cp850][i]], "%c", i);
    }
  }
  /* unfortunately this is only true for part of cp437 */
  for (i=0; i<256; i++) {
    sprintf(from_latin1_table[cp437][i], "?");
  }
  for (i=0; i<256; i++) {
    if (to_latin1_table[cp437][i] != '?') {
      sprintf(from_latin1_table[cp437][to_latin1_table[cp437][i]], "%c", i);
    }
  }
  /* now for the missing characters... */
  /* " ":       EI    */
  sprintf(from_latin1_table[cp437][0xcb], "\"E");
  sprintf(from_latin1_table[cp437][0xcf], "\"I");
  /* ' ':    yA  IOUY */
  sprintf(from_latin1_table[cp437][0xfd], "'y");
  sprintf(from_latin1_table[cp437][0xc1], "'A");
  sprintf(from_latin1_table[cp437][0xcd], "'I");
  sprintf(from_latin1_table[cp437][0xd3], "'O");
  sprintf(from_latin1_table[cp437][0xda], "'U");
  sprintf(from_latin1_table[cp437][0xdd], "'Y");
  /* ^ ^:     A EIOU  */
  sprintf(from_latin1_table[cp437][0xc2], "^A");
  sprintf(from_latin1_table[cp437][0xca], "^E");
  sprintf(from_latin1_table[cp437][0xce], "^I");
  sprintf(from_latin1_table[cp437][0xd4], "^O");
  sprintf(from_latin1_table[cp437][0xdb], "^U");
  /* ` `:     A EIOU  */ 
  sprintf(from_latin1_table[cp437][0xc0], "`A");
  sprintf(from_latin1_table[cp437][0xc8], "`E");
  sprintf(from_latin1_table[cp437][0xcc], "`I");
  sprintf(from_latin1_table[cp437][0xd2], "`O");
  sprintf(from_latin1_table[cp437][0xd9], "`U");
  /* ~ ~: a o A   O   */
  sprintf(from_latin1_table[cp437][0xe3], "~a");
  sprintf(from_latin1_table[cp437][0xf5], "~o");
  sprintf(from_latin1_table[cp437][0xc3], "~A");
  sprintf(from_latin1_table[cp437][0xd5], "~O");
  /* / /:  do  D  O   */
  sprintf(from_latin1_table[cp437][0xf0], "/d");
  sprintf(from_latin1_table[cp437][0xf8], "/o");
  sprintf(from_latin1_table[cp437][0xd0], "/D");
  sprintf(from_latin1_table[cp437][0xd8], "/O");
  
  /* ascii is even worse: it needs all of the above, and more */
  for (i=0; i<128; i++) {
    sprintf(from_latin1_table[ascii][i], "%s", from_latin1_table[cp437][i]);
  }
  /* Now for the missing characters */
  sprintf(from_latin1_table[ascii][0xE1], "'a");           /* ' ' */
  sprintf(from_latin1_table[ascii][0xE9], "'e");
  sprintf(from_latin1_table[ascii][0xED], "'i");
  sprintf(from_latin1_table[ascii][0xF3], "'o");
  sprintf(from_latin1_table[ascii][0xFA], "'u");
  sprintf(from_latin1_table[ascii][0xFD], "'y");
  sprintf(from_latin1_table[ascii][0xC1], "'A");
  sprintf(from_latin1_table[ascii][0xC9], "'E");
  sprintf(from_latin1_table[ascii][0xCD], "'I");
  sprintf(from_latin1_table[ascii][0xD3], "'O");
  sprintf(from_latin1_table[ascii][0xDA], "'U");
  sprintf(from_latin1_table[ascii][0xDD], "'Y");
  sprintf(from_latin1_table[ascii][0xE6], "@a");           /* @ @ */
  sprintf(from_latin1_table[ascii][0xC6], "@A");
  sprintf(from_latin1_table[ascii][0xE2], "^a");           /* ^ ^ */
  sprintf(from_latin1_table[ascii][0xEA], "^e");
  sprintf(from_latin1_table[ascii][0xEE], "^i");
  sprintf(from_latin1_table[ascii][0xF4], "^o");
  sprintf(from_latin1_table[ascii][0xFB], "^u");
  sprintf(from_latin1_table[ascii][0xC2], "^A");
  sprintf(from_latin1_table[ascii][0xCA], "^E");
  sprintf(from_latin1_table[ascii][0xCE], "^I");
  sprintf(from_latin1_table[ascii][0xD4], "^O");
  sprintf(from_latin1_table[ascii][0xDB], "^U");
  sprintf(from_latin1_table[ascii][0xE7], ",c");           /* , , */
  sprintf(from_latin1_table[ascii][0xC7], ",C");
  sprintf(from_latin1_table[ascii][0xE0], "`a");           /* ` ` */
  sprintf(from_latin1_table[ascii][0xE8], "`e");
  sprintf(from_latin1_table[ascii][0xEC], "`i");
  sprintf(from_latin1_table[ascii][0xF2], "`o");
  sprintf(from_latin1_table[ascii][0xF9], "`u");
  sprintf(from_latin1_table[ascii][0xC0], "`A");
  sprintf(from_latin1_table[ascii][0xC8], "`E");
  sprintf(from_latin1_table[ascii][0xCC], "`I");
  sprintf(from_latin1_table[ascii][0xD2], "`O");
  sprintf(from_latin1_table[ascii][0xD9], "`U");
  sprintf(from_latin1_table[ascii][0xE4], "\"a");           /* " " */
  sprintf(from_latin1_table[ascii][0xEB], "\"e");
  sprintf(from_latin1_table[ascii][0xEF], "\"i");
  sprintf(from_latin1_table[ascii][0xF6], "\"o");
  sprintf(from_latin1_table[ascii][0xDF], "\"s");
  sprintf(from_latin1_table[ascii][0xFC], "\"u");
  sprintf(from_latin1_table[ascii][0xFF], "\"y");
  sprintf(from_latin1_table[ascii][0xC4], "\"A");
  sprintf(from_latin1_table[ascii][0xCB], "\"E");
  sprintf(from_latin1_table[ascii][0xCF], "\"I");
  sprintf(from_latin1_table[ascii][0xD6], "\"O");
  sprintf(from_latin1_table[ascii][0xDC], "\"U");
  sprintf(from_latin1_table[ascii][0xE5], "*a");           /* * * */
  sprintf(from_latin1_table[ascii][0xC5], "*A");
  sprintf(from_latin1_table[ascii][0xE3], "~a");           /* ~ ~ */
  sprintf(from_latin1_table[ascii][0xF1], "~n");
  sprintf(from_latin1_table[ascii][0xF5], "~o");
  sprintf(from_latin1_table[ascii][0xC3], "~A");
  sprintf(from_latin1_table[ascii][0xD1], "~N");
  sprintf(from_latin1_table[ascii][0xD5], "~O");
  sprintf(from_latin1_table[ascii][0xF0], "/d");           /* / / */
  sprintf(from_latin1_table[ascii][0xF8], "/o");
  sprintf(from_latin1_table[ascii][0xD0], "/D");
  sprintf(from_latin1_table[ascii][0xD8], "/O");
  sprintf(from_latin1_table[ascii][0xBF], "/?");
  sprintf(from_latin1_table[ascii][0xA1], "/!");
}

/* recode special characters:
   "a =  (and similar for eiouyAEIOU  --- cp437 only aeiouyAOU)
   "s =  
   'a =  (and similar for eiouyAEIOUY --- cp437 only aeiouE)
   ^a =  (and similar for eiouAEIOU   --- cp437 only aeiou)
   `a =  (and similar for eiouAEIOU   --- cp437 only aeiou)
   *a =  (and similar for A)
   ,c =  (and similar for C)
   ~a = a (and similar for noANO       --- cp437 only nN)
   @a =  (and similar for A)
   /o = o (and similar for O           --- not in cp437)
   /d = ? (and similar for D           --- not in cp437)
   /? =  (and similar for !)

   This is currently implemented for cp437, cp850 and iso-latin-1.

   To keep Emacs happy:
*/

static unsigned char acute[3][256] = {{0}};/* ' ' */
static unsigned char at[3][256]    = {{0}};/* @ @ */
static unsigned char circu[3][256] = {{0}};/* ^ ^ */
static unsigned char comma[3][256] = {{0}};/* , , */
static unsigned char grave[3][256] = {{0}};/* ` ` */
static unsigned char quote[3][256] = {{0}};/* " " */
static unsigned char star[3][256]  = {{0}};/* * * */
static unsigned char tilde[3][256] = {{0}};/* ~ ~ */
static unsigned char slash[3][256] = {{0}};/* / / */

void init_recode(void)
{
  int i,j;			       /* loop control */

  for (j=0; j<3; j++) {
    for (i=0; i<256; i++) {
      /* self-insert command */
      acute[j][i] = i;                 /* ' ' */
      at[j][i]    = i;                 /* @ @ */
      circu[j][i] = i;                 /* ^ ^ */
      comma[j][i] = i;                 /* , , */
      grave[j][i] = i;                 /* ` ` */
      quote[j][i] = i;                 /* " " */
      star[j][i]  = i;                 /* * * */
      tilde[j][i] = i;                 /* ~ ~ */
      slash[j][i] = i;                 /* / / */
    } /* end for i */
  } /* end for j */
  /* exceptions: */
  /* E,I are missing in cp437 */
  acute[cp437]['a'] = 0xA0;            /* ' ' */
  acute[cp437]['e'] = 0x82;
  acute[cp437]['i'] = 0xA1;
  acute[cp437]['o'] = 0xA2;
  acute[cp437]['u'] = 0xA3;
  acute[cp437]['E'] = 0x90;
  at[cp437]['a']    = 0x91;            /* @ @ */
  at[cp437]['A']    = 0x92;
  /* A,E,I,O,U are missing in cp437 */
  circu[cp437]['a'] = 0x83;            /* ^ ^ */
  circu[cp437]['e'] = 0x88;
  circu[cp437]['i'] = 0x8C;
  circu[cp437]['o'] = 0x93;
  circu[cp437]['u'] = 0x96;
  comma[cp437]['C'] = 0x80;
  comma[cp437]['c'] = 0x87;
  /* A,E,I,O,U are missing in cp437 */
  grave[cp437]['a'] = 0x85;            /* ` ` */
  grave[cp437]['e'] = 0x8A;
  grave[cp437]['i'] = 0x8D;
  grave[cp437]['o'] = 0x95;
  grave[cp437]['u'] = 0x97;
  /* E,I are missing in cp437 */
  quote[cp437]['a'] = 0x84;            /* " " */
  quote[cp437]['e'] = 0x89;
  quote[cp437]['i'] = 0x8B;
  quote[cp437]['o'] = 0x94;
  quote[cp437]['u'] = 0x81;
  quote[cp437]['y'] = 0x98;
  quote[cp437]['A'] = 0x8E;
  quote[cp437]['O'] = 0x99;
  quote[cp437]['U'] = 0x9A;
  quote[cp437]['s'] = 0xE1;
  star[cp437]['a']  = 0x86;            /* * * */
  star[cp437]['A']  = 0x8F;
  /* a,A,o,O are missing in cp437 */
  tilde[cp437]['n'] = 0xA4;            /* ~ ~ */
  tilde[cp437]['N'] = 0xA5;
  /* d,D (thorn) and o,O (danish OE) are missing in cp437 */
  slash[cp437]['?'] = 0xA8;            /* / / */
  slash[cp437]['!'] = 0xAD;

  acute[cp850]['a'] = 0xA0;            /* ' ' */
  acute[cp850]['e'] = 0x82;
  acute[cp850]['i'] = 0xA1;
  acute[cp850]['o'] = 0xA2;
  acute[cp850]['u'] = 0xA3;
  acute[cp850]['y'] = 0xEC;
  acute[cp850]['A'] = 0xB5;
  acute[cp850]['E'] = 0x90;
  acute[cp850]['I'] = 0xD6;
  acute[cp850]['O'] = 0xE0;
  acute[cp850]['U'] = 0xE9;
  acute[cp850]['Y'] = 0xED;
  at[cp850]   ['a'] = 0x91;            /* @ @ */
  at[cp850]   ['A'] = 0x92;
  circu[cp850]['a'] = 0x83;            /* ^ ^ */
  circu[cp850]['e'] = 0x88;
  circu[cp850]['i'] = 0x8C;
  circu[cp850]['o'] = 0x93;
  circu[cp850]['u'] = 0x96;
  circu[cp850]['A'] = 0xB6;
  circu[cp850]['E'] = 0xD2;
  circu[cp850]['I'] = 0xD7;
  circu[cp850]['O'] = 0xE2;
  circu[cp850]['U'] = 0xEA;
  comma[cp850]['c'] = 0x87;            /* , , */
  comma[cp850]['C'] = 0x80;
  grave[cp850]['a'] = 0x85;            /* ` ` */
  grave[cp850]['e'] = 0x8A;
  grave[cp850]['i'] = 0x8D;
  grave[cp850]['o'] = 0x95;
  grave[cp850]['u'] = 0x97;
  grave[cp850]['A'] = 0xB7;
  grave[cp850]['E'] = 0xD4;
  grave[cp850]['I'] = 0xDE;
  grave[cp850]['O'] = 0xE3;
  grave[cp850]['U'] = 0xEB;
  quote[cp850]['a'] = 0x84;            /* " " */
  quote[cp850]['e'] = 0x89;
  quote[cp850]['i'] = 0x8B;
  quote[cp850]['o'] = 0x94;
  quote[cp850]['s'] = 0xE1;
  quote[cp850]['u'] = 0x81;
  quote[cp850]['y'] = 0x98;
  quote[cp850]['A'] = 0x8E;
  quote[cp850]['E'] = 0xD3;
  quote[cp850]['I'] = 0xD8;
  quote[cp850]['O'] = 0x99;
  quote[cp850]['U'] = 0x9A;
  star[cp850] ['a']  = 0x86;           /* * * */
  star[cp850] ['A']  = 0x8F;
  tilde[cp850]['a'] = 0xC6;            /* ~ ~ */
  tilde[cp850]['n'] = 0xA4;
  tilde[cp850]['o'] = 0xE4;
  tilde[cp850]['A'] = 0xC7;
  tilde[cp850]['N'] = 0xA5;
  tilde[cp850]['O'] = 0xE5;
  slash[cp850]['d'] = 0xD0;            /* / / */
  slash[cp850]['o'] = 0x9B;
  slash[cp850]['D'] = 0xD1;
  slash[cp850]['O'] = 0x9D;
  slash[cp850]['?'] = 0xA8;
  slash[cp850]['!'] = 0xAD;

  acute[latin1]['a'] = 0xE1;           /* ' ' */
  acute[latin1]['e'] = 0xE9;
  acute[latin1]['i'] = 0xED;
  acute[latin1]['o'] = 0xF3;
  acute[latin1]['u'] = 0xFA;
  acute[latin1]['y'] = 0xFD;
  acute[latin1]['A'] = 0xC1;
  acute[latin1]['E'] = 0xC9;
  acute[latin1]['I'] = 0xCD;
  acute[latin1]['O'] = 0xD3;
  acute[latin1]['U'] = 0xDA;
  acute[latin1]['Y'] = 0xDD;
  at[latin1]   ['a'] = 0xE6;           /* @ @ */
  at[latin1]   ['A'] = 0xC6;
  circu[latin1]['a'] = 0xE2;           /* ^ ^ */
  circu[latin1]['e'] = 0xEA;
  circu[latin1]['i'] = 0xEE;
  circu[latin1]['o'] = 0xF4;
  circu[latin1]['u'] = 0xFB;
  circu[latin1]['A'] = 0xC2;
  circu[latin1]['E'] = 0xCA;
  circu[latin1]['I'] = 0xCE;
  circu[latin1]['O'] = 0xD4;
  circu[latin1]['U'] = 0xDB;
  comma[latin1]['c'] = 0xE7;           /* , , */
  comma[latin1]['C'] = 0xC7;
  grave[latin1]['a'] = 0xE0;           /* ` ` */
  grave[latin1]['e'] = 0xE8;
  grave[latin1]['i'] = 0xEC;
  grave[latin1]['o'] = 0xF2;
  grave[latin1]['u'] = 0xF9;
  grave[latin1]['A'] = 0xC0;
  grave[latin1]['E'] = 0xC8;
  grave[latin1]['I'] = 0xCC;
  grave[latin1]['O'] = 0xD2;
  grave[latin1]['U'] = 0xD9;
  quote[latin1]['a'] = 0xE4;           /* " " */
  quote[latin1]['e'] = 0xEB;
  quote[latin1]['i'] = 0xEF;
  quote[latin1]['o'] = 0xF6;
  quote[latin1]['s'] = 0xDF;
  quote[latin1]['u'] = 0xFC;
  quote[latin1]['y'] = 0xFF;
  quote[latin1]['A'] = 0xC4;
  quote[latin1]['E'] = 0xCB;
  quote[latin1]['I'] = 0xCF;
  quote[latin1]['O'] = 0xD6;
  quote[latin1]['U'] = 0xDC;
  star[latin1] ['a'] = 0xE5;           /* * * */
  star[latin1] ['A'] = 0xC5;
  tilde[latin1]['a'] = 0xE3;           /* ~ ~ */
  tilde[latin1]['n'] = 0xF1;
  tilde[latin1]['o'] = 0xF5;
  tilde[latin1]['A'] = 0xC3;
  tilde[latin1]['N'] = 0xD1;
  tilde[latin1]['O'] = 0xD5;
  slash[latin1]['d'] = 0xF0;           /* / / */
  slash[latin1]['o'] = 0xF8;
  slash[latin1]['D'] = 0xD0;
  slash[latin1]['O'] = 0xD8;
  slash[latin1]['?'] = 0xBF;
  slash[latin1]['!'] = 0xA1;
}

char* recode(char answer[],           /* array to recode (in place) */
             int n,                    /* max size of array */
             int the_coding_system    /* target coding */
             )
{
  int m;
  unsigned char (*tab)[3][256] = NULL;
  int i,j;

  if (slash[latin1]['a'] == 0) {
    init_recode();
  }
  m = strlen(answer);
  if (m >= n) {
    fprintf(stderr, "String to long in recode(), (m=%d) >= (n=%d)\n", m,n);
    answer[n]=0;
    m = n-1;
  }
  if (the_coding_system == ascii) {
    /* nothing to do */
    return answer;
  }

  for (i=0,j=0; i<m; ) {
    switch (answer[i]) {
    case '\'':
      tab = &acute;
      break;
    case '@':
      tab = &at;
      break;
    case '^':
      tab = &circu;
      break;
    case ',':
      tab = &comma;
      break;
    case '`':
      tab = &grave;
      break;
    case '"':
      tab = &quote;
      break;
    case '*':
      tab = &star;
      break;
    case '~':
      tab = &tilde;
      break;
    case '/':
      tab = &slash;
      break;
    case '\\':
      /* apply default action to next character */
      /* copy backlash */
      answer[j++] = answer[i++];
      answer[j++] = answer[i++];
      break;
    default:
      answer[j++] = answer[i++];
      break;
    } /* end switch */
    if (tab != NULL) {
      if ((*tab)[the_coding_system][(int)answer[i+1]] != answer[i+1]) {
        answer[j++] = (*tab)[the_coding_system][(int)answer[i+1]];
        i += 2;
        tab = NULL;
      } else {
        answer[j++] = answer[i++];
        tab = NULL;
      }
    } /* end tab != NULL */
  } /* end for i */
  answer[j] = 0;
  
  return answer;
}
