/* I'm not really sure this struct should ever appear outside
   quizcard.c, but for now it seems easiest to do so... */

#ifndef QUIZCARD_H
#define QUIZCARD_H

/* system-specific stuff.  This should go into a header-file of it's
   own, and possibly even into the makefile ... */
typedef unsigned char byte;
#ifdef unix
typedef unsigned short word;
#else
typedef unsigned int word;
#endif


/* general definitions: */
#define FALSE 0
#define TRUE  1

#define EMPTY ""

#define MAX_NUM_ENTRIES 2048

/* The first byte of each entry codes quite a bit of Information:

   LWWWGGST

   where the letters are:
   L   = Language, 0 or 1.
   WWW = Wordform: Noun=0, Verb=1, Adjective=2, Adverb=3, Preposition=4, Other=7
   GG  = Gender: Male=0, Female=1, Neutral=2
   S   = Singular=0, Plural=1
   T   = Tested during the current run
*/

#define LANGUAGE 128
#define LANG0    0
#define LANG1    128
#define WORDFORM (64+32+16)
#define NOUN     0
#define VERB     (1<<4)
#define ADJECTIV (2<<4)
#define ADVERB   (3<<4)
#define PREPOSIT (4<<4)
#define OTHER    (7<<4)
#define GENDER   (8+4)
#define MALE     0
#define FEMALE   (1<<2)
#define NEUTER   (2<<2)
#define CASE     2
#define SINGULAR 0
#define PLURAL   (1<<1)
#define TESTED   1

#define MAX_PHRASE_LEN 100

typedef struct entry_t{
  word number;                         /* unique id */
  byte byte1;                          /* flags, s.a. */
  char *phrase;                        /* The actual phrase */
  byte num_translations;               /* the number of translations */
  word *translations;                  /* one or more translations */
  byte *count;                         /* the number of succesfull tries for each translation */
} entry;

char readkey(char* chars);

#endif
